@@ -25,7 +25,7 @@
 
 #include <glib.h>
 #include <gtk/gtk.h>
-#include <libgnomeui/gnome-desktop-thumbnail.h>
+#include <libgnome-desktop/gnome-desktop-thumbnail.h>
 #include <string.h>
 
 #include "cheese-fileutil.h"
@@ -50,6 +50,8 @@ typedef struct
   GnomeDesktopThumbnailFactory *factory;
   gboolean multiplex_thumbnail_generator;
   guint n_items;
+  guint idle_id;
+  GQueue *thumbnails;
 } CheeseThumbViewPrivate;
 
 enum
@@ -76,14 +78,25 @@ typedef struct
   CheeseThumbView *thumb_view;
   GFile *file;
   GtkTreeIter iter;
-} CheeseThumbViewThreadData;
+} CheeseThumbViewIdleData;
+
 
 static void
-cheese_thumb_view_thread_append_item (gpointer data)
+cheese_thumb_view_constructed (GObject *object);
+
+static gboolean
+cheese_thumb_view_idle_append_item (gpointer data)
 {
-  CheeseThumbViewThreadData *item       = data;
-  CheeseThumbView           *thumb_view = item->thumb_view;
-  CheeseThumbViewPrivate    *priv       = CHEESE_THUMB_VIEW_GET_PRIVATE (thumb_view);
+  CheeseThumbViewIdleData *item = g_queue_peek_head (data);
+  CheeseThumbView         *thumb_view;
+  CheeseThumbViewPrivate  *priv;
+
+  /* Disconnect the idle handler when the queue is empty. */
+  if (item == NULL) return FALSE;
+
+  thumb_view = item->thumb_view;
+  priv = CHEESE_THUMB_VIEW_GET_PRIVATE (thumb_view);
+
 
   GnomeDesktopThumbnailFactory *factory = priv->factory;
   GFile                        *file    = item->file;
@@ -101,7 +114,7 @@ cheese_thumb_view_thread_append_item (gpointer data)
   if (!info)
   {
     g_warning ("Invalid filename\n");
-    return;
+    return TRUE;
   }
   g_file_info_get_modification_time (info, &mtime);
   mime_type = g_strdup (g_file_info_get_content_type (info));
@@ -149,7 +162,7 @@ cheese_thumb_view_thread_append_item (gpointer data)
     if (error)
     {
       g_warning ("%s", error->message);
-      return;
+      return TRUE;
     }
   }
   else
@@ -157,18 +170,17 @@ cheese_thumb_view_thread_append_item (gpointer data)
     cheese_thumbnail_add_frame (&pixbuf);
   }
 
-  gdk_threads_enter ();
-
   gtk_list_store_set (priv->store, &iter,
                       THUMBNAIL_PIXBUF_COLUMN, pixbuf, -1);
 
-  gdk_threads_leave ();
-
   g_free (mime_type);
   g_free (filename);
   g_object_unref (pixbuf);
   g_object_unref (file);
-  g_free (item);
+  g_slice_free (CheeseThumbViewIdleData, item);
+  g_queue_pop_head (data);
+
+  return TRUE;
 }
 
 static void
@@ -184,11 +196,11 @@ cheese_thumb_view_append_item (CheeseThumbView *thumb_view, GFile *file)
   GError       *error = NULL;
   gboolean      skip  = FALSE;
 
-  CheeseThumbViewThreadData *data;
+  CheeseThumbViewIdleData *data;
 
   filename = g_file_get_path (file);
 
-  if (!(g_str_has_suffix (filename, PHOTO_NAME_SUFFIX)) && !(g_str_has_suffix (filename, VIDEO_NAME_SUFFIX)))
+  if (!(g_str_has_suffix (filename, CHEESE_PHOTO_NAME_SUFFIX)) && !(g_str_has_suffix (filename, CHEESE_VIDEO_NAME_SUFFIX)))
   {
     g_free (filename);
     return;
@@ -263,18 +275,13 @@ cheese_thumb_view_append_item (CheeseThumbView *thumb_view, GFile *file)
 
   if (!priv->multiplex_thumbnail_generator)
   {
-    data             = g_new0 (CheeseThumbViewThreadData, 1);
+    data             = g_slice_new0 (CheeseThumbViewIdleData);
     data->thumb_view = g_object_ref (thumb_view);
     data->file       = g_object_ref (file);
     data->iter       = iter;
 
-    if (!g_thread_create ((GThreadFunc) cheese_thumb_view_thread_append_item,
-                          data, FALSE, &error))
-    {
-      g_error ("Failed to create thumbnail thread: %s\n", error->message);
-      g_error_free (error);
-      return;
-    }
+    g_queue_push_tail (priv->thumbnails, data);
+    if (!priv->idle_id) g_idle_add (cheese_thumb_view_idle_append_item, priv->thumbnails);
   }
 }
 
@@ -484,7 +491,7 @@ cheese_thumb_view_fill (CheeseThumbView *thumb_view)
   CheeseThumbViewPrivate *priv = CHEESE_THUMB_VIEW_GET_PRIVATE (thumb_view);
 
   GDir       *dir_videos, *dir_photos;
-  char       *path_videos, *path_photos;
+  const char *path_videos, *path_photos;
   const char *name;
   char       *filename;
   GFile      *file;
@@ -511,7 +518,7 @@ cheese_thumb_view_fill (CheeseThumbView *thumb_view)
     /* read videos from the vid directory */
     while ((name = g_dir_read_name (dir_videos)))
     {
-      if (!(g_str_has_suffix (name, VIDEO_NAME_SUFFIX)))
+      if (!(g_str_has_suffix (name, CHEESE_VIDEO_NAME_SUFFIX)))
         continue;
 
       filename = g_build_filename (path_videos, name, NULL);
@@ -529,7 +536,7 @@ cheese_thumb_view_fill (CheeseThumbView *thumb_view)
     /* read photos from the photo directory */
     while ((name = g_dir_read_name (dir_photos)))
     {
-      if (!(g_str_has_suffix (name, PHOTO_NAME_SUFFIX)))
+      if (!(g_str_has_suffix (name, CHEESE_PHOTO_NAME_SUFFIX)))
         continue;
 
       filename = g_build_filename (path_photos, name, NULL);
@@ -554,6 +561,7 @@ cheese_thumb_view_finalize (GObject *object)
   g_object_unref (priv->factory);
   g_file_monitor_cancel (priv->photo_file_monitor);
   g_file_monitor_cancel (priv->video_file_monitor);
+  g_queue_free (priv->thumbnails);
 
   G_OBJECT_CLASS (cheese_thumb_view_parent_class)->finalize (object);
 }
@@ -563,6 +571,7 @@ cheese_thumb_view_class_init (CheeseThumbViewClass *klass)
 {
   GObjectClass *object_class = G_OBJECT_CLASS (klass);
 
+  object_class->constructed = cheese_thumb_view_constructed;
   object_class->finalize = cheese_thumb_view_finalize;
 
   g_type_class_add_private (klass, sizeof (CheeseThumbViewPrivate));
@@ -606,6 +615,12 @@ cheese_thumb_view_init (CheeseThumbView *thumb_view)
 
   priv->store   = gtk_list_store_new (3, GDK_TYPE_PIXBUF, G_TYPE_STRING, G_TYPE_STRING);
   priv->n_items = 0;
+  priv->idle_id = 0;
+  priv->thumbnails = g_queue_new ();
+  
+  priv->fileutil = cheese_fileutil_new ();
+  priv->factory = gnome_desktop_thumbnail_factory_new (GNOME_DESKTOP_THUMBNAIL_SIZE_NORMAL);
+
 
   g_signal_connect (G_OBJECT (priv->store),
                     "row-inserted",
@@ -616,8 +631,21 @@ cheese_thumb_view_init (CheeseThumbView *thumb_view)
                     G_CALLBACK (cheese_thumb_view_row_deleted_cb),
                     thumb_view);
 
-  priv->fileutil = cheese_fileutil_new ();
+  g_signal_connect (G_OBJECT (thumb_view), "drag-data-get",
+                    G_CALLBACK (cheese_thumb_view_on_drag_data_get_cb), NULL);
+                    
+  /* We do the rest of the initialization in our constructed() implementation,
+   * because GtkIconView may not be ready for us to do more now.
+   * See https://bugzilla.gnome.org/show_bug.cgi?id=643286#c6
+   */
+}
 
+static void
+cheese_thumb_view_constructed (GObject *object)
+{
+  CheeseThumbView *thumb_view = CHEESE_THUMB_VIEW (object);
+  CheeseThumbViewPrivate *priv = CHEESE_THUMB_VIEW_GET_PRIVATE (thumb_view);
+  
   gtk_icon_view_set_model (GTK_ICON_VIEW (thumb_view), GTK_TREE_MODEL (priv->store));
 
   gtk_widget_set_size_request (GTK_WIDGET (thumb_view),
@@ -628,8 +656,6 @@ cheese_thumb_view_init (CheeseThumbView *thumb_view)
   gtk_icon_view_set_row_spacing (GTK_ICON_VIEW (thumb_view), 0);
   gtk_icon_view_set_column_spacing (GTK_ICON_VIEW (thumb_view), 0);
 
-  priv->factory = gnome_desktop_thumbnail_factory_new (GNOME_DESKTOP_THUMBNAIL_SIZE_NORMAL);
-
   gtk_icon_view_set_pixbuf_column (GTK_ICON_VIEW (thumb_view), 0);
 
   gtk_icon_view_set_columns (GTK_ICON_VIEW (thumb_view), G_MAXINT);
@@ -638,12 +664,10 @@ cheese_thumb_view_init (CheeseThumbView *thumb_view)
                                           target_table, G_N_ELEMENTS (target_table),
                                           GDK_ACTION_COPY);
   gtk_icon_view_set_selection_mode (GTK_ICON_VIEW (thumb_view), GTK_SELECTION_MULTIPLE);
-  g_signal_connect (G_OBJECT (thumb_view), "drag-data-get",
-                    G_CALLBACK (cheese_thumb_view_on_drag_data_get_cb), NULL);
 
   gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE (priv->store),
                                         THUMBNAIL_BASENAME_URL_COLUMN, GTK_SORT_ASCENDING);
-
+                                        
   cheese_thumb_view_fill (thumb_view);
 }
 
@@ -657,7 +681,7 @@ cheese_thumb_view_new ()
 }
 
 void
-cheese_thumb_view_start_monitoring_photo_path (CheeseThumbView *thumb_view, char *path_photos)
+cheese_thumb_view_start_monitoring_photo_path (CheeseThumbView *thumb_view, const char *path_photos)
 {
   CheeseThumbViewPrivate *priv = CHEESE_THUMB_VIEW_GET_PRIVATE (thumb_view);
 
@@ -676,7 +700,7 @@ cheese_thumb_view_start_monitoring_photo_path (CheeseThumbView *thumb_view, char
 }
 
 void
-cheese_thumb_view_start_monitoring_video_path (CheeseThumbView *thumb_view, char *path_videos)
+cheese_thumb_view_start_monitoring_video_path (CheeseThumbView *thumb_view, const char *path_videos)
 {
   CheeseThumbViewPrivate *priv = CHEESE_THUMB_VIEW_GET_PRIVATE (thumb_view);
 
